#!/usr/bin/perl
#
# Matthew J. Kidd (San Diego, CA)
#
# This software is released under the GNU General Public License GPLv3
# See: http://www.gnu.org/licenses/gpl.html for full license.
#
# Calculates the partnership payoff matrix by directly scanning ACBLscore game files.
# Also generates player and partnership statistics.
#
# Dependencies: JSON, Date-DayOfWeek. JSON is shipped with most Perl installations.
# Run 'ppm install Date-DayOfWeek' to install Date-DayOfWeek.
#
# 24-Oct-2013 - Current release
# 05-Oct-2013 - Initial release
#
# See also ACBLmerge (http://lajollabridge.com/Software/ACBLmerge/ACBLmergeAbout.htm)

use strict;

use Date::DayOfWeek;
use JSON;

my $VERSTR = '1.0.1';

my $DAYS = {'sun' => 0, 'mon' => 1, 'tue' => 2, 'wed' => 3, 'thu' => 4,
  'fri' => 5, 'sat' => 6};

if ( scalar(@ARGV) == 0 ) {
  
	(my $bname = $0) =~ s/.*[\\\/]//;
	
  print "\n  Missing parameter(s)\n";
  
  print << "DONE";

  Usage: $bname [fname...] [-d dirname] [-pl plfname] [-pr prfname]
         [-pm pmfname] [-js jsfname] [-s list] [-dy list] [-y list]
         [-ul list] [-e list] [-ms #] [-v #] [-vr]

  fname  : Input filename(s), i.e. ACBLscore game files
  dname  : Directory name (processes all games files in the directory)
           The -d option may be repeated to specify multiple directories

  Output files
    -pl : List of players and their statistics
    -pr : List of partnerships and their statistics
    -pm : Payoff matrix data (partnership-partnership interaction)
    -js : JavaScript variable assignment useful for HTML matrix display
    
  General options
    -v  : Set verbosity level. 0 -> Quiet mode (minimize command line output)
          2 -> Detailed. Default is 1. Meaningful up to 4.
    -vr : Show version
    
  Game file filtering options
    -s  : List of sessions (M = morning, A = afternoon, E = evening, L = late)
          to examine, e.g. AE. If not specified, all sessions are examined.
    -dy : Comma separated list of days to examine: mon, tue, wed, thu,
          fri, sat, sun. If not specified, all days are examined.
    -y  : Comma separated list of years to examine. Ranges may also be
          specified, e.g. 2007-2009,2011
          
  Event filtering options
    -ul : Comma separated list of masterpoint upper limits for events
          to examine. Use 'open' or 0 for events with no upper limit.       
    -e  : Comma separated list of event indices to examine. For example,
          if game files always have an open event and may optionally have
          a limited event as the second event then -e 1 will limit the
          examination to the open event. -ul is usually the better switch.
          
  Result filtering options
    -ms : Minimum number of sessions that a pair must play in to be reported
          in partnership list (-pr) and payoff matrix (-pm) output.    

  Computes the partnership payoff matrix by directly scanning ACBLscore game
  files. Also computes player and partnership statistics.

  Online documentation is located at:
  http://www.lajollabridge.com/Software/Payoff-Matrix/Payoff-Matrix-About.htm

DONE
  exit;
}

my (@fnames, %opt, $errmsg);

while ( my $arg = shift(@ARGV) ) {
  if ( substr($arg,0,1) ne '-' ) { push @fnames, $arg; next; };
  my $sw = substr($arg,1);
  if ($sw eq 'd') {
    $arg = shift(@ARGV);
    if (!defined $arg) {
      print STDERR "\n-d switch is missing argument.\n"; exit;
    }
    push @{$opt{'dirnames'}}, $arg;
  }
  elsif ($sw eq 'dy') {
    $arg = shift(@ARGV);
    if (!defined $arg) {
      print STDERR "\n-dy switch is missing argument.\n"; exit;
    }
    $opt{'days'} = [0, 0, 0, 0, 0, 0, 0];
    foreach my $day (split /,/, $arg) {
      (my $lday = $day) =~ tr /A-Z/a-z/;
      if (!defined $DAYS->{$lday}) {
        print STDERR "\nBad day for -dy argument: $lday (use sun, mon, tue, wed, thu, fri, sat).\n";
        exit;
      }
      else {
        $opt{'days'}[$DAYS->{$lday}] = 1;
      }
    }
  }
  elsif ($sw eq 'e') {
    $arg = shift(@ARGV);
    if (!defined $arg) {
      print STDERR "\n-$sw switch is missing argument.\n"; exit;
    }
    my @ixlist = split(',', $arg);
    foreach my $ix (@ixlist) {
      if ($ix !~ /^\d+$/) {
        print STDERR "\nAll indices for -$sw switch must be an integers. Bad value: $ix\n"; exit;
      }
      $opt{'eidx'}{$ix} = 1;
    }
  }
  elsif ($sw eq 'ul') {
    $arg = shift(@ARGV);
    if (!defined $arg) {
      print STDERR "\n-$sw switch is missing argument.\n"; exit;
    }
    my @ixlist = split(',', $arg);
    foreach my $ix (@ixlist) {
      if ($ix ne 'open' && $ix !~ /^\d+$/) {
        print STDERR "\nAll values for -$sw switch must be integers or 'open'. Bad value: $ix\n"; exit;
      }
      $opt{'ul'}{$ix} = $ix eq 'open' ? 0 : $ix;
    }
  }
  elsif ($sw eq 'pl' || $sw eq 'pr' || $sw eq 'pm' || $sw eq 'js') {
    # Handle all the output filenames.
    $arg = shift(@ARGV);
    if (!defined $arg) {
      print STDERR "\n-$sw switch is missing argument.\n"; exit;
    }
    $opt{'plfname'} = $arg if $sw eq 'pl';
    $opt{'prfname'} = $arg if $sw eq 'pr';
    $opt{'pmfname'} = $arg if $sw eq 'pm';
    $opt{'jsfname'} = $arg if $sw eq 'js';
  }
  elsif ($sw eq 'v') {
    $arg = shift(@ARGV);
    if (!defined $arg) {
      print STDERR "\n-$sw switch is missing argument.\n"; exit;
    }    
    if ($arg !~ /^\d+$/ ) {
      print STDERR "\n-$sw switch requires an integer argument.\n"; exit;
    }
    $opt{'verbosity'} = $arg;
  }
  elsif ($sw eq 'vr')  { $opt{'showver'} = 1; }
  elsif ($sw eq 'ms') {
    $arg = shift(@ARGV);
    if (!defined $arg) {
      print STDERR "\n-$sw switch is missing argument.\n"; exit;
    }    
    if ($arg !~ /^\d+$/ ) {
      print STDERR "\n-$sw switch requires an integer argument.\n"; exit;
    }
    $opt{'minSessions'} = $arg;
  }
  elsif ($sw eq 's') {
    $arg = shift(@ARGV);
    if (!defined $arg) {
      print STDERR "\n-$sw switch is missing argument.\n"; exit;
    }
    if ($arg !~ /^[MAEL]+$/i) {
      print STDERR "\n-sw argument can only contain these letters: M, A, E, and L. Bad value: $arg\n"; exit;
    }
    ($opt{'sessions'} = $arg) =~ tr/a-z/A-Z/;
  }
  elsif ($sw eq 'y') {
    $arg = shift(@ARGV);
    if (!defined $arg) {
      print STDERR "\n-$sw switch is missing argument.\n"; exit;
    }
    foreach my $yearspan (split /,/, $arg) {
      if ( $yearspan =~ /^\d+$/ )  { $opt{'years'}{$yearspan} = undef; }
      elsif ( $yearspan =~ /^\d+-\d+$/ ) {
        my ($year1, $year2) = split /-/, $yearspan;
        for (my $year=$year1; $year <= $year2; $year++) {
          $opt{'years'}{$year} = undef;
        }
      }
      else {
        print STDERR "\nEach -$sw option list value must be an integer or an integer range.\n";
        print STDERR "$yearspan list value was invalid in list: $arg\n";
        exit;
      }
    }
  }
  else {
    print STDERR "Unrecognized switch: $arg\n";
  }
}

# Use default when value is not specified.
$opt{'verbosity'} = 1 if !defined $opt{'verbosity'};

print STDERR "Payoff Matrix $VERSTR\n" if $opt{'showver'};

if ($opt{'dirnames'}) {
  foreach my $dirname ( @{$opt{'dirnames'}} ) {
    if (! -d $dirname) {
      print STDERR "Nonexistant folder: $dirname (skipping)\n"; next;
    }
    opendir DH, $dirname or die "Unable to open directory: $dirname";
    my @dfnames = grep { -f "$dirname/$_" } readdir(DH);
    closedir(DH);
    foreach my $fname (@dfnames) { push @fnames, "$dirname\\$fname"; }
  }
}


# $PL contains player data/statistics, $PR contains pair data/statistics.
# $PM contains payoff matrix data/statistics. $PNN maps player names to
# player numbers - allows in part for handling players who do not have
# an ACBL player number.
my ($pl, $pr, $pm, $pnn); my $pnnIndex = 0;
my $ngames = 0; my $highest_monthnum; my $nonPairSections = 0;

# Loop through and process all relevant ACBLscore game files.
foreach my $fname (@fnames) {
  if ($fname !~ /\.ac(m|a|e|l|b)$/i) { next; }
  (my $lastchar = substr($fname,-1)) =~ tr /a-z/A-Z/;
  if ( $opt{'sessions'} && index($opt{'sessions'}, $lastchar) == -1) { next; }

  (my $bname = $fname) =~ s/.*[\\\/]//;
  my $year = substr($bname,0,2);
  if ($year !~ /\d\d/) { $year = undef; }
  elsif ($year >= 80) { $year += 1900; }
  else { $year += 2000; }
  
  if ( $opt{'years'} && !exists $opt{'years'}{$year} ) { next; }
    
  if ( $opt{'days'} ) {
    if (!defined $year) { next; }
    my $dow = dayofweek( substr($bname,4,2), substr($bname,2,2), $year );
    if (! $opt{'days'}[$dow] ) { next; }
  }
  my $monthnum = 0;
  if ($bname =~ /^\d{4}/) { $monthnum = $year * 12 + substr($bname,2,2); }
  $highest_monthnum = $monthnum if $monthnum > $highest_monthnum;
  
  my $errmsg = parseACBLgamefile($fname, $highest_monthnum);
  if ($errmsg) { print STDERR $errmsg, "\n"; next; }
  $ngames++; 
}

my $nplayers1 = scalar(keys %$pl);
my $npairs1   = scalar(keys %$pr);
my $nmatrix1  = scalar(keys %$pm);

# If -ms option is specified, trim $PR to pairs meeting minimum number of
# sessions and $PM to interaction between pairs meeting the criterion. 
($pr, $pm) = trim($pr, $pm) if defined $opt{'minSessions'};

$pl = meanstd($pl);
$pr = meanstd($pr);
$pm = meanstd($pm);

# Assign rank index (1..N) to each partnership.
ixAssign();

if ( defined $opt{'plfname'} ) {
  $errmsg = dumpPlayers($opt{'plfname'}, $pl);
  print STDERR "Player dump failed: $errmsg\n" if $errmsg;
} 
if ( defined $opt{'prfname'} ) {
  $errmsg = dumpPartnerships($opt{'prfname'}, $pr, $pl);
  print STDERR "Partnership dump failed: $errmsg\n" if $errmsg;
}
if ( defined $opt{'pmfname'} ) {
  $errmsg = dumpPayoffMatrix($opt{'pmfname'}, $pm, $pr);
   print STDERR "Payoff matrix dump failed: $errmsg\n" if $errmsg;
}
if ( defined $opt{'jsfname'} ) {
  $errmsg = dumpJSON($opt{'jsfname'});
   print STDERR "JSON dump failed: $errmsg\n" if $errmsg;
}

if ( $opt{'verbosity'} ) {
  print STDERR "Game files processed: $ngames\n";
  
  if ($ngames > 0) {
    # Note: Some pairs can not interact because they have a player in common. The
    # percentage calculation does not account for this, though it does account for
    # the fact that a partnership can not interact with itself.
    my $maxInteractions1 = $npairs1 * ($npairs1 - 1) / 2;
    print STDERR  "Total unique players: ", $nplayers1, "\n";
    print STDERR  "Total partnerships across all files and sessions: ", $npairs1, "\n";
    printf STDERR "Total interacting pairs of partnerships: %d  (%0.2f%%)\n",
      $nmatrix1, $npairs1 ? 100 * $nmatrix1 / $maxInteractions1 : 0;
      
    if (defined $opt{'minSessions'}) {
      print STDERR "\nAfter imposing $opt{'minSessions'} session minimum cut\n";
      
      my $nplayers2 = scalar(keys %$pl);
      my $npairs2   = scalar(keys %$pr);
      my $nmatrix2  = scalar(keys %$pm);
      
      my $maxInteractions2 = $npairs2 * ($npairs2 - 1) / 2;
      print STDERR  "Total unique players: ", $nplayers2, "\n";
      print STDERR  "Total partnerships across all files and sessions: ", $npairs2, "\n";
      printf STDERR "Total interacting pairs of partnerships: %d  (%0.2f%%)\n",
      $nmatrix2, $npairs2 ? 100 * $nmatrix2 / $maxInteractions2 : 0;
    }
  }
  
  if ($nonPairSections > 0) {
    print "Total non-pair sections skipped: $nonPairSections\n";
  }
}

sub parseACBLgamefile {
  # Parse binary ACBLscore output. This format isn't documented but I have partly
  # reverse engineered it for pairs games.
  
  if (scalar(@_) != 2) { die "function requires 2 arguments."; }
  my ($fname, $highest_monthnum) = @_;
  
  (my $bname = $fname) =~ s/.*[\\\/]//;
  
  my ($data, $errmsg, $pnum);
  # Safely isolate $/ change.
  {
    local $/; my $fh;
    return("Unable to open/read ACBLscore gamefile: $fname") if !open($fh, '<', $fname);
    binmode($fh);
    $data = <$fh>;
    close($fh);
  }
  
  # Check for the 'AC3' magic bytes that seem to appear at the start of all ACBLscore
  # game files.
  if (substr($data,0,6) ne "\x12\x0a\x03AC3") {
    return("Does not appear to be an ACBLscore game file: $fname");
  }
  
  print STDERR "Processing game file: $fname\n" if $opt{'verbosity'} > 1;
  
  my @sections;
  # List of sections begins at offset 0x13e and each section is 22 bytes long.
  my $ix = 0x13e;
  while ( ord(substr($data,$ix,1)) ) {
    my ($eventnum, $zLetter, $pDetails, $pBoards, $mp100, $secnum, $nplayed1, $nplayed2) =
      unpack('CC x2 VVv x C x3 CC', substr($data,$ix,22) );
    my $pEvent = unpack('V', substr($data,14+4*$eventnum));
    my $mpLimit = unpack('v', substr($data,$pEvent+0xf4,2));
    push @sections, {'eventnum' => $eventnum, 'letter' => substr($data,$ix+2,$zLetter),
      'pDetails' => $pDetails, 'pBoards' => $pBoards, 'mp100' => $mp100, 'secnum' => $secnum,
      'nplayed1' => $nplayed1, 'nplayed2' => $nplayed2, 'mpLimit', $mpLimit };
    $ix += 22;
  }
  
  if ($opt{'verbosity'} > 1) {
    print STDERR "List of sections\n";
    for my $i (0..$#sections) {
      my $sc = $sections[$i];
      my $mpStr = $sc->{'mpLimit'} ? $sc->{'mpLimit'} : 'Open';
      printf STDERR "%2s %2d %2d %7.2f %2d %2d %4s  Section Details at 0x%04x  Board Index at 0x%04x\n", 
        $sc->{'letter'}, $sc->{'eventnum'}, $sc->{'secnum'}, $sc->{'mp100'}, $sc->{'nplayed1'},
        $sc->{'nplayed2'}, $mpStr, $sc->{'pDetails'}, $sc->{'pBoards'};
    }
    print STDERR "\n";
  }
  
  # Process each section.
  for my $i (0..$#sections) {
    my $sc = $sections[$i];
    # Skip section if it is not a pairs section.
    if ( !$sc->{'pBoards'} ) { $nonPairSections++; next; }
    
    # Skip section if it does not have a desired masterpoint upper limit.
    next if defined $opt{'ul'} && !$opt{'ul'}{$sc->{'mpLimit'}};
    
    # Skip section if it belongs to an event number which is to be skipped.
    next if defined $opt{'eidx'} && !$opt{'eidx'}{$sc->{'eventnum'}};
    
    # $PS (as in "Pairs in Section")
    my ($errmsg, $isHowell, $ps) = parseSectionPairs($data, $sc->{'pDetails'}, $bname, $sc->{'letter'});
    if ($errmsg) {
      print STDERR "Problem with section $sc->{'letter'} of $bname: $errmsg (skipping)\n";
      next;
    }
 
    # Create unique identifier for each partnership by combining player numbers.
    # Also update parternship and player statistics
    foreach my $j (0..$#$ps) {
      my ($pnum1, $pnum2, $pair_id, $fl, $pct2); 
      while ( my($pairnum, $p) = each $ps->[$j] ) {
        # Assign pair_id of the form 'pnum1+pnum2'
        $pnum1 = $p->{'pnum'}[0]; $pnum2 = $p->{'pnum'}[1];
        $pair_id = $pnum1 < $pnum2 ? $pnum1 . '+' . $pnum2 : $pnum2 . '+' . $pnum1;
        $p->{'pair_id'} = $pair_id;
        
        # Update pair statistics.
        if ( !defined $pr->{$pair_id} ) {
          $fl = $pnum1 < $pnum2 ? 0 : 1;
          $pr->{$pair_id} = {
            'pnum1' => $p->{'pnum'}[$fl], 'pnum2' => $p->{'pnum'}[1-$fl],
            'fullname1' => $p->{'fname'}[$fl] . ' ' . $p->{'lname'}[$fl],
            'fullname2' => $p->{'fname'}[1-$fl] . ' ' . $p->{'lname'}[1-$fl],
            'mpsum' => 0, 'cnt' => 0, 'pctsum' => 0, 'pctsum2' => 0,
            'pctmax' => 0, 'slamcnt' => 0, 'rulecnt' => 0
          };
        }
        $pct2 = $p->{'pct'} * $p->{'pct'};
        $pr->{$pair_id}{'cnt'}++;
        $pr->{$pair_id}{'pctsum'}  += $p->{'pct'};
        $pr->{$pair_id}{'pctsum2'} += $pct2;
        $pr->{$pair_id}{'mpsum'}   += $p->{'mp'};
        $pr->{$pair_id}{'pctmax'} = $p->{'pct'} if $pr->{$pair_id}{'pctmax'} < $p->{'pct'};
      
        # Update player statistics.
        for my $k (0..1) {
          if ( !defined $pl->{$p->{'pnum'}[$k]} ) {
            $pl->{$p->{'pnum'}[$k]} = {
              'fname' => $p->{'fname'}[$k], 'lname' => $p->{'lname'}[$k],
              'fullname' => $p->{'fname'}[$k] . ' ' . $p->{'lname'}[$k],
              'mpsum' => 0, 'cnt' => 0, 'pctsum' => 0, 'pctsum2' => 0,
              'pctmax' => 0
            };
          }
          $pnum = $p->{'pnum'}[$k];
          $pl->{$pnum}{'cnt'}++;
          $pl->{$pnum}{'pctsum'}  += $p->{'pct'};
          $pl->{$pnum}{'pctsum2'} += $pct2;
          $pl->{$pnum}{'mpsum'}   += $p->{'mp'};
          $pl->{$pnum}{'pctmax'} = $p->{'pct'} if $pl->{$pnum}{'pctmax'} < $p->{'pct'};
        }
      }
    }
    
    # Parse board results, updating $PM payoff matrix
    parseSectionBoards($data, $sc->{'pBoards'}, $ps, $isHowell, $bname, $sc->{'letter'});
  }
  
  return undef;
}

sub parseSectionPairs {
  # Parse information about pairs in a section.
  if (scalar(@_) != 4) { die "function requires 4 arguments."; }
  my ($data, $ix, $bname, $secLetter) = @_;
  
  my $PAIR_ENTRY_LEN = 0x192;
  
  my @ps; 
  my ($len, $iy);

  # The Section Details Table contains at least two, though possibly up to
  # four, pointers to Pair Index Tables. The two Pair Index Tables list
  # which pairs are seated N-S and E-W respectively at each table at round
  # one. Note: It is important to use the Pair Index Tables because although
  # the Pair entries are usually consecutive, this is not always true. One
  # exception occurs when a pair is added to a full Mitchell movement after
  # one round has been entered. (See 121125A U526 unit game, Pair 17 N-S in
  # section A.)
  
  my @pPairIndexTables = unpack('V2', substr($data,$ix+4,8));
  my @unk = unpack('V2', substr($data,$ix+12,8));
  if ( ($unk[0] || $unk[1]) and $opt{'verbosity'} ) {
    printf STDERR "Warning: Spare values in the Section Details Table that might be used as\n" .
      "additional Pair Table pointers are not both zero (0x%04x and 0x%04x). This\n" .
      "may indicate a movement that is not handled properly by this program.\n\n", 
      $unk[0], $unk[1];
  }
  
  my $pMovement = unpack('V', substr($data,$ix+20,4));
  my ($nTables, @tables, $tablenum);
  
  # Use the Movement table to figure out which pair numnber was seated N-S
  # and E-W at each table at round one. For Mitchell movements this is an
  # unnecessary extra step but it is the general technique. For Howell
  # movements it is essential.
  $nTables = unpack('C', substr($data, $pMovement+4, 1));
  if ( $opt{'verbosity'} > 3 ) {
    printf STDERR "%d tables in Movement Table at 0x%04x\n", $nTables, $pMovement;
  }
  
  # The Movement Table has a 7 byte header and 75 bytes per table. The 75 bytes
  # represent up to 25 rounds where each group of three bytes gives the N-S pair
  # number at the table, the E-W pair number at the table, and the lowest board
  # number at the table for the round.
  $tables[0] = [ unpack("(C x74)$nTables",   substr($data, $pMovement+7, $nTables * 75)) ];
  $tables[1] = [ unpack("(x C x73)$nTables", substr($data, $pMovement+7, $nTables * 75)) ];

  # For a Howell movement, the pair numbers for the pairs seated N-S and E-W at
  # round 1 will not overlap.
  my $isHowell = 1;
  my @allpairnums = sort( ( @{$tables[0]}, @{$tables[1]} ) );
  for (my $i=0; $i<$#allpairnums-1; $i++) {
    if ( $allpairnums[$i] == $allpairnums[$i+1] ) { $isHowell = 0; last; }
  }

  if ( $opt{'verbosity'} > 3 ) {
    printf STDERR "\n %s Table:", $isHowell ? '  (Howell)' : '(Mitchell)';
    for my $i (1..$nTables) { printf STDERR " %2d", $i; }
    printf STDERR "\n", ;
    for my $i (0..1) {
      print STDERR "Round 1 ", $i ? "E-W" : "N-S", " pairs:";
      foreach my $pairnum (@{$tables[$i]}) { printf STDERR " %2d", $pairnum; }
      if ($i) { print STDERR "\n\n"; } else { print STDERR "\n"; } 
    }
  }
  
  
  my ($pairnum, $tot, $pct, $mp, $SA, $OA, $SB, $OB, $SC, $OC,
      @zlname, @zfname, @zcity, @zstate, @zpnum,
      @zlname, @zfname, @zcity, @zstate, @zpnum);
  my (@lname, @fname, @city, @state, @pnum);  
  
  my $nPairsRecorded = 0;
  for my $idir (0..1) {
    my $psIndex = $isHowell ? 0 : $idir;
    my $ip = $pPairIndexTables[$idir];
    if ($ip == 0) {
      printf STDERR "Warning: Pair Index Table %d seems to be missing.\n", $idir+1;
    };
    
    # This number includes phantom pairs.
    my $nPairs = unpack('C', substr($data, $ip+6, 1));
    if ( $opt{'verbosity'} > 3 ) {
      printf STDERR "%d pairs in Pair Index Table at 0x%04x\n", $nPairs, $ip;    
    }
    
    # Skip to first Pair Index Entry and advance 4 more bytes to location of pointer.
    $ip += 24;
    for (my $k=0; $k < $nPairs; $k++) {
      $ix = unpack('V', substr($data,$ip,4));
      $ip += 8;
      
      $len = unpack('v', substr($data,$ix,2));
      if ($len != $PAIR_ENTRY_LEN) { 
        printf STDERR "Pair Table Entry %d of section %s does not have expected " . 
          "length (%d) in file\n$bname. (Fatal error)\n", $k+1, $secLetter, $PAIR_ENTRY_LEN;
        exit;
      }

      # Each pair record has 162 bytes of pair data (beyond the record length) and two player
      # fields of 120 bytes = 402 bytes total (0x192).
      ($tablenum, $tot, $pct, $mp, $SA, $OA, $SB, $OB, $SC, $OC,
       $zlname[0], $zfname[0], $zcity[0], $zstate[0], $zpnum[0],
       $zlname[1], $zfname[1], $zcity[1], $zstate[1], $zpnum[1]) =
      unpack('x2 C x9 v x14 v x46 v x16 (C x3 C x15)3 x10 (C x16 C x16 C x16 C x2 C x7 x58)2',
        substr($data, $ix, $PAIR_ENTRY_LEN + 2) );
      $tot /= 100; $mp /= 100; $pct /= 100;
      
      # Convert table number to pair number using the Movement table data.
      if (!$isHowell && $tablenum > $nTables) {
        # This case is seen in 120427.ACA La Jolla unit game. This is a bit of kludge because
        # I don't know exactly how this case is treated in ACBLscore.
        print STDERR "A pair seems to be have been added to table $secLetter$tablenum of a Mitchell movement\n" .
          "after the first round in game file $bname. Will assume pair number\n" . 
          "is equal to the table number.\n\n" if $opt{'verbosity'};
        $pairnum = $tablenum;
      }
      else {
        $pairnum = $tables[$idir][$tablenum-1];        
      }
      
      # Check to avoid phantom pairs
      if ($zfname[$0] != 0 || $zlname[0] != 0 || $zfname[1] != 0 || $zlname[1] != 0) {
        # Grab strings, canonicalize life masters and assign artificial player numbers to non-ACBL members.
        for my $i (0..1) {
          $nPairsRecorded++;
          
          $iy = $ix + ($i ? 0x11c : 0xa4);
          $lname[$i] = substr($data, $iy+1,  $zlname[$i]);
          $fname[$i] = substr($data, $iy+18, $zfname[$i]);
          $city [$i] = substr($data, $iy+35, $zcity [$i]);
          $state[$i] = substr($data, $iy+52, $zstate[$i]);
          $pnum [$i] = substr($data, $iy+55, $zpnum [$i]);        
    
          if ( $pnum[$i] =~ /^[1-9J-R][0-9]{6}$/ ) {
            # Canonicalize life masters.
            $pnum[$i] =~ tr/J-R/1-9/;            
          }
          else {
            if ($pnum[$i] ne 'NM' && $zpnum[$i] > 0) {
              # 'NM' is sometimes used for 'Non Member'. Anything else is weird. The 090219.ACE
              # Soledad Club file has '434900' for Ed Kennedy and the HTML report says 'BAD #'
              # underneath the player number. Maybe the ACBLscore input validation was not as
              # good back then. Warn the user but still assign an artificial player number.
              print STDERR "Bad player number: $pnum[$i] in file $bname. ",
                "(artificial number assigned)\n" if $opt{'verbosity'};
            }
            # Assign an artificial player number if we have not already done so.
            my $fullname = $fname[$i] . ' ' . $lname[$i];
            $pnn->{$fullname} = ++$pnnIndex if !defined $pnn->{$fullname};
            $pnum[$i] = $pnn->{$fullname};
          }

          if ($opt{'verbosity'} > 1) {
            printf STDERR "%2d %-30s %7d  %6.2f %5.2f %5s  %d %d %d  %d %d %d\n",
              $pairnum, $fname[$i] . ' ' . $lname[$i], $pnum[$i], 
              $tot, $pct, $mp ? sprintf('%5.2f', $mp) : '', $OA, $OB, $OC, $SA, $SB, $SC;
          }
          
        }
        print STDERR "\n" if $opt{'verbosity'} > 1;
        
        $ps[$psIndex]{$pairnum} = {'tot' => $tot, 'pct' => $pct, 'mp' => $mp, 'SA' => $SA, 'OA' => $OA,
          'SB' => $SB, 'OB' => $OB, 'SC' => $SC, 'OC' => $OC, 'lname' => [@lname], 'fname' => [@fname],
          'city' => [@city], 'state' => [@state], 'pnum' => [@pnum]};
      }      

    }
  }
  
  # This error can occur if we are processing an event that was setup names were never entered.
  my $errmsg = $nPairsRecorded ? undef : "No pairs found";
  return ($errmsg, $isHowell, \@ps);
}


sub parseSectionBoards {
  # Parse information about boards in a section.
  if (scalar(@_) != 6) { die "function requires 6 arguments."; }
  my ($data, $ix, $ps, $isHowell, $bname, $secLetter) = @_;

  my $verbose = $opt{'verbosity'} > 2;
  my $len = unpack('v', substr($data,$ix,2));
  my $nboards = unpack('C', substr($data,$ix+4));
  
  # Start array with common, i.e. undoubled slam scores first, NT, majors,
  # and then minors. Implementing a binary search or using a hash to define
  # the value score would be the normal approach but a linear search with
  # the most common slam scores packed at the front of the array will be
  # just as quick for this particular case.
  # 
  # Note: 920 is also 4m**+0 V, 4m**+1 NV, 1M**+2 NV, 4N**+0 NV.
  # 940 is also 1m*+4 V. 980 is also 1N*+4 V, 2m*+5 V. 990 is also 2N*+5 NV
  # 4M*+1 V. 1010 is also 4M*+1 V. 1090 is also 2N*+2 V. 1190 is also
  # 4M*+2 V. 1210 is also 4M*+2 V. 1239 is also 1m*+5 NV. 1310 is also
  # 4m*+3 V. 1330 is also 3M*+3 V. 1380 is also 1N*+6 V. 1390 is also
  # 4H*+3 V. 1430 is also 1m**+3 V and 1m**+6 NV. 1440 is also 2M**+4 NV.
  # 1460 is also 2M*+4 V. 1520 is also 1M**+2 V. 1680 is also 2N**+2 V.
  # 1830 is also 1m**+4 V. 1880 is also 4M**+2 V. 1960 is also 1N**+3 V.
  # 2230 is also 1m**+5 V. 2280 is also 4M**+3 V.
  
  my @slam_scores = (990, 980, 1020, 1010, 920, 940, 1370, 1390, 1430, 1440,
    1470, 2140, 2210, 2220, 1090, 1190, 1210, 1230, 1310, 1330, 1380, 1390,
    1460, 1510, 1520, 1540, 1580, 1620, 1630, 1660, 1680, 1740, 1770, 1790,
    1820, 1830, 1860, 1880, 1960, 2070, 2110, 2230, 2240, 2280, 2330, 2470,
    2490, 2510, 2660, 2940, 2980);
  
  print STDERR "Number of boards: $nboards\n" if $verbose;
  
  # 38 bytes gets to the first board index entry. File offset is four bytes in. 
  $ix += 42;
  
  my($iy, $blen, $bnum, $nrounds, $mp_top);
  my($round, $ns_pair, $ew_pair, $ns_score, $ew_score, $ns_mp, $ew_mp);
  my($ns_pair_id, $ew_pair_id, $pct, $pair_id1, $pair_id2, $pair_pair_id);

  for (my $nb = 0; $nb < $nboards; $nb++) {
    # Each board has results for every round. If a board was not in play on a
    # given round of the movement or it met a phantom pair, special scores
    # indicate this situation.
    $iy = unpack('V', substr($data,$ix,4));
    ($blen, $bnum, $nrounds) = unpack('vC x2 C', substr($data,$iy,6) );
    if ($nrounds * 6 + 18 == $blen) {
      die "$blen / $nrounds mismatch.  blen: $blen  nrounds: $nrounds";
    }
    # Skip to the first result for the board.
    $iy += 6;
    
    print STDERR "\nBoard $bnum\n" if $verbose;
    for (my $nr = 0; $nr < $nrounds; $nr++ ) {
      # Score unpacks as 1/10 of actual score, i.e. superfluous zero is dropped
      # The ACBLscore interface for entering scores does the same thing. The
      # matchpoint scores are 100 times the actual value. Note: we have both
      # N-S and E-W scores because director rulings can result in split scores.
      # Also the byte before the N-S pair number seems to always match the N-S
      # pair number for a Mitchell movement but is different for a Howell movement.
      ($round, $ns_pair, $ns_score, $ns_mp, $ew_pair, $ew_score, $ew_mp) =
        unpack('C x (C x s< v x2)2', substr($data,$iy,18) );
      $iy += 18;  
      
      if ($verbose) {
        printf STDERR "%2d  %2d vs %2d %5d %5d %4.2f %4.2f\n", $round, $ns_pair, $ew_pair,
          $ns_score * 10, $ew_score * 10, $ns_mp / 100, $ew_mp / 100;
      }
      
      # Board was not in play on this round.
      next if ($ns_pair == 0 || $ew_pair == 0);
      
      $ns_pair_id = $ps->[0]{$ns_pair}{'pair_id'};
      $ew_pair_id = $ps->[$isHowell ? 0 : 1]{$ew_pair}{'pair_id'};
      
      # Pair lookup can fail if the Pair Table for the section was not read correctly.
      # This can happen in weird cases, e.g. when a pair is added after a round is
      # already played such that the pairs in the Pair Table are not all contiguous.
      if (!defined $ns_pair_id || !defined $ew_pair_id) {
        print STDERR "Pair lookup failure: $bname $secLetter:", 
        defined $ns_pair_id ? '' : " N-S $ns_pair",
        defined $ew_pair_id ? '' : " E-W $ew_pair", "\n";
        next;
      }
           
      # Ignore split scores (they result from director rulings), Avg, Avg+, and Avg-
      # results. 7600 (down 13 redoubled vul), stored as 760, is the best/worst score
      # a pair can recieve. Values higher than this are special (950 => Not Played,
      # 999 => Placeholder (board not in played on round), 2040 => Ave-, 2050 => Ave,
      # 2060 => Ave+).
      if ( ($ns_score + $ew_score) != 0 || $ns_score > 760 || $ew_score > 760) {
        # 999 check is probably redundant with ($ns_pair == 0 || $ew_pair == 0) check
        if ($ns_score != 950 && $ns_score != 999) {
          # Keep track of which partnerships cause director rulings.
          $pr->{$ns_pair_id}{'rulecnt'}++; $pr->{$ew_pair_id}{'rulecnt'}++;
        }
        next;
      }

      # It doesn't make sense to compute this for every result but one must be careful
      # about using the number of times the board is play due to scaling when sections
      # are combined and boards are played an unequal number of times.
      $mp_top = ($ns_mp + $ew_mp) / 100;
      
      if ($ns_pair_id le $ew_pair_id) {
        $pair_id1 = $ns_pair_id; $pair_id2 = $ew_pair_id;
        $pct = $ns_mp / $mp_top;
      }
      else {
        $pair_id1 = $ew_pair_id; $pair_id2 = $ns_pair_id;
        $pct = $ew_mp / $mp_top;
      }
            
      $pair_pair_id = $pair_id1 . ':' . $pair_id2;
      if ( !defined $pm->{$pair_pair_id} ) {
        $pm->{$pair_pair_id} = { 'cnt' => 1, 'pctsum' => $pct, 'pctsum2' => $pct * $pct };
      }
      else {
        $pm->{$pair_pair_id}{'cnt'}++;
        $pm->{$pair_pair_id}{'pctsum'}  += $pct;
        $pm->{$pair_pair_id}{'pctsum2'} += $pct * $pct;
      }
      
      # Credit partnerships with any slams bid and made. No slam scores are divisible
      # by 100, so check for this immediately to avoid scanning the entire array for
      # for scores from defeated doubled or redoubled contracts.
      if ( abs($ns_score) >= 92 && $ns_score % 10 != 0 ) {
        # Check if it is possible slam score.
        my $isSlam; my $abs_score = abs($ns_score) * 10;
        foreach my $slam_score (@slam_scores) {
          if ( $abs_score == $slam_score ) { $isSlam = 1; last; }
        }
        if ($isSlam) {
          if ($ns_score > 0) { $pr->{$ns_pair_id}{'slamcnt'}++; }
          else { $pr->{$ew_pair_id}{'slamcnt'}++; }
        } 
      }
      
    }
    $ix += 8;
  } 


}

sub trim {
  if (scalar(@_) != 2) { die "function requires 2 arguments."; }
  my ($pr, $pm) = @_;
  
  my ($pair_id1, $pair_id2);
  foreach my $k (keys %$pr) { delete $pr->{$k} if $pr->{$k}{'cnt'} < $opt{'minSessions'}; }
  foreach my $k (keys %$pm) {
    ($pair_id1, $pair_id2) = split /\:/, $k;
    delete $pm->{$k} if !defined $pr->{$pair_id1} || !defined $pr->{$pair_id2};
  }
  
  return ($pr, $pm);
}

sub ixAssign {
  my $ix = 0;
  my @keys = sort {$pr->{$b}{'pct'} <=> $pr->{$a}{'pct'} } keys %$pr;
  foreach my $k (@keys) { $ix++; $pr->{$k}{'ix'} = $ix; }
}

sub dumpPlayers {
  # Dump player information sorted by decreasing average percentage.
  if (scalar(@_) != 2) { die "function requires 2 arguments."; }
  my ($fname, $pl) = @_;
  
  return "Unable to open/write: $fname" if ! open(my $fh, '>', $fname);

  my @hrow = ('pnum', 'fname', 'lname', 'fullname', 'cnt', 'pct', 'meanerr', 'std', 
    'pctmax', 'mpsum', 'mpavg');
  print $fh join("\t", @hrow), "\n";
  
  my (@row, $pct, $meanerr, $std, $mpavg, $pctmax);
  my @keys = sort {$pl->{$b}{'pct'} <=> $pl->{$a}{'pct'} } keys %$pl;
  foreach my $k (@keys) {
    $pct = sprintf('%.2f', $pl->{$k}{'pct'});
    $std = defined $pl->{$k}{'std'} ? sprintf('%.2f', $pl->{$k}{'std'}) : '';
    $meanerr = defined $pl->{$k}{'meanerr'} ? sprintf('%.2f', $pl->{$k}{'meanerr'}) : '';
    $mpavg = sprintf('%.2f', $pl->{$k}{'mpsum'} / $pl->{$k}{'cnt'});
    $pctmax = sprintf('%.2f', $pl->{$k}{'pctmax'});
    
    @row = ( $k, $pl->{$k}{'fname'}, $pl->{$k}{'lname'}, $pl->{$k}{'fullname'},
      $pl->{$k}{'cnt'}, $pct, $meanerr, $std, $pctmax, $pl->{$k}{'mpsum'}, $mpavg );
    print $fh join("\t", @row), "\n";
  }
  
  close($fh);
  return undef;
}

sub dumpPartnerships {
  # Dump partnership information sorted by decreasing average percentage.
  if (scalar(@_) != 3) { die "function requires 3 arguments."; }
  my ($fname, $pr, $pl) = @_;
  
  return "Unable to open/write: $fname" if ! open(my $fh, '>', $fname);
  
  my @hrow = ('ix', 'pnum1', 'pnum2', 'fullname1', 'fullname2', 'cnt', 'pct',
    'meanerr', 'std', 'pctmax', 'mpsum', 'mpavg', 'slamcnt', 'slamavg', 'rulecnt');
  print $fh join("\t", @hrow), "\n";
  
  my (@row, $pnum1, $pnum2, $pct, $meanerr, $std, $pctmax, $mpavg, $slamavg);
  my @keys = sort {$pr->{$a}{'ix'} <=> $pr->{$b}{'ix'} } keys %$pr;
  foreach my $k (@keys) {
    ($pnum1, $pnum2) = split /\+/, $k;
    $pct = sprintf('%.2f', $pr->{$k}{'pct'});
    $std = defined $pr->{$k}{'std'} ? sprintf('%.2f', $pr->{$k}{'std'}) : '';
    $meanerr = defined $pr->{$k}{'meanerr'} ? sprintf('%.2f', $pr->{$k}{'meanerr'}) : '';
    $pctmax = sprintf('%.2f', $pr->{$k}{'pctmax'});
    $mpavg = sprintf('%.2f', $pr->{$k}{'mpsum'} / $pr->{$k}{'cnt'});
    $slamavg = sprintf('%.2f', $pr->{$k}{'slamcnt'} / $pr->{$k}{'cnt'});
        
    @row = ($pr->{$k}{'ix'}, $pnum1, $pnum2, $pl->{$pnum1}{'fullname'}, $pl->{$pnum2}{'fullname'},
      $pr->{$k}{'cnt'}, $pct, $meanerr, $std, $pctmax, $pr->{$k}{'mpsum'}, $mpavg,
      $pr->{$k}{'slamcnt'}, $slamavg, $pr->{$k}{'rulecnt'} );
    print $fh join("\t", @row), "\n";
  }
  
  close($fh);
  return undef;
}

sub dumpPayoffMatrix {
  # Dump payoff matrix sorted by decreasing disparity between partnerships.
  if (scalar(@_) != 3) { die "function requires 3 arguments."; }
  my ($fname, $pm, $pr) = @_;
  
  return "Unable to open/write: $fname" if ! open(my $fh, '>', $fname);

  my @hrow = ('ix1', 'ix2', 'pair_id1', 'pair_id2', 'pair_name1', 'pair_name2', 'cnt', 'pct', 
    'meanerr', 'std', 'disparity', 'relperf');
  print $fh join("\t", @hrow), "\n";
  
  my (@row, $pair_id1, $pair_id2, $pair_name1, $pair_name2, $pct, $std, $meanerr,
    $disparity, $relperf);
  my @keys = sort { abs(50 - $pm->{$b}{'pct'}) <=> abs(50 - $pm->{$a}{'pct'}) } keys %$pm;
  foreach my $k (@keys) {
    ($pair_id1, $pair_id2) = split /\:/, $k;
    $pair_name1 = $pr->{$pair_id1}{'fullname1'} . ' - ' . $pr->{$pair_id1}{'fullname2'};
    $pair_name2 = $pr->{$pair_id2}{'fullname1'} . ' - ' . $pr->{$pair_id2}{'fullname2'};
    
    $pct = sprintf('%.2f', $pm->{$k}{'pct'});
    $std = defined $pm->{$k}{'std'} ? sprintf('%.2f', $pm->{$k}{'std'}) : '';
    $meanerr = defined $pm->{$k}{'meanerr'} ? sprintf('%.2f', $pm->{$k}{'meanerr'}) : '';
    $disparity = sprintf('%.2f', 2 * abs(50 - $pm->{$k}{'pct'}) );
    $relperf = sprintf('%.2f', 2*$pm->{$k}{'pct'} -100 - $pr->{$pair_id1}{'pct'} + $pr->{$pair_id2}{'pct'});
    
    @row = ($pr->{$pair_id1}{'ix'}, $pr->{$pair_id2}{'ix'}, $pair_id1, $pair_id2,
      $pair_name1, $pair_name2, $pm->{$k}{'cnt'}, $pct, $meanerr, $std, $disparity, $relperf);
    print $fh join("\t", @row), "\n";
  }
  
  close($fh);
  return undef;
}

sub dumpJSON {
  if (scalar(@_) != 1) { die "function requires 1 argument."; }
  my ($fname) = @_;
  
  # Pack data in a way that minimizes amount of output, i.e. as a bunch of arrays of the same
  # length instead of an array of hashes, for faster web page loading.
  my ($pnum1, $pnum2, $fd, $pj, $pair_id1, $pair_id2, $pct, $meanerr, $relperf);
  my $pairs = {'fullname1' => [], 'fullname2' => [], 'cnt' => [], 'pct' => [], 'std' => [], 'meanerr' => []};
  my @keys = sort {$pr->{$a}{'ix'} <=> $pr->{$b}{'ix'} } keys %$pr;
  foreach my $k (@keys) {
    ($pnum1, $pnum2) = split /\+/, $k;
    push @{$pairs->{'fullname1'}}, $pl->{$pnum1}{'fullname'};
    push @{$pairs->{'fullname2'}}, $pl->{$pnum2}{'fullname'};
    push @{$pairs->{'cnt'}}, $pr->{$k}{'cnt'};
    foreach my $fd ('pct', 'std', 'meanerr') {
      push @{$pairs->{$fd}}, int(100 * $pr->{$k}{$fd} + 0.5) / 100;
    }
  }
  
  # Dump Payoff Matrix as a sparse matrix.
  while (my($k, $v) = each %$pm) {
    ($pair_id1, $pair_id2) = split /\:/, $k;
    $pct = int(100 * $v->{'pct'} + 0.5) / 100;
    $meanerr = int(100 * $v->{'meanerr'} + 0.5) / 100;
    $relperf = 2*$pm->{$k}{'pct'} -100 - $pr->{$pair_id1}{'pct'} + $pr->{$pair_id2}{'pct'};
    $relperf = int(100 * $relperf + 0.5) / 100;
    
    # Zero based pair numbers for convenience in Javascript code.  
    $pj->{ $pr->{$pair_id1}{'ix'}-1 }{ $pr->{$pair_id2}{'ix'}-1 } = 
      [ $v->{'cnt'}, $pct, $meanerr, $relperf ];
  }
  
  my $data = {'pairs' => $pairs, 'pm' => $pj};

  return "Unable to open/write: $fname" if ! open(my $fh, '>', $fname);
  
  (my $bname = $fname) =~ s/.*[\\\/]//;
  $bname =~ s/\.[^.]*$//;
  $bname =~ s/-payoff-matrix$//;
  
  printf $fh "payoffMatrix[\"%s\"] = ", $bname;
  print $fh JSON->new->encode($data);
  print $fh ';';

  close($fh);
  return undef;
}

sub meanstd {
  # Calculate average percentage, standard deviation, and error on average percentage.
  if (scalar(@_) != 1) { die "function requires 1 argument."; }
  my ($x) = @_;
  
  my $std2;
  while ( my ($key, $p) = each %$x ) {
    if ($p->{'cnt'} == 0) { use Data::Dumper; print Dumper($p); exit; } ##
    $p->{'pct'} = $p->{'pctsum'} / $p->{'cnt'};
    $std2 = ($p->{'pctsum2'} - $p->{'pctsum'} * $p->{'pctsum'} / $p->{'cnt'}) / ($p->{'cnt'}-1)
      if $p->{'cnt'} > 1;
    # If all data have the same value, roundoff error can cause the calculation
    # of the standard deviation to fail. Handle this but still fail if STD2 is too
    # negative since that would indicate a problem upstream.  
    $std2 = 0 if $std2 < 0 && $std2 > -1e-10; 
    $p->{'std'} = $p->{'cnt'} < 2 ? undef : sqrt($std2);
    $p->{'meanerr'} = $p->{'cnt'} < 2 ? undef : $p->{'std'} / sqrt($p->{'cnt'});
  }
  return $x;
}
