function varargout = payoffmatrixim(arg1, arg2, imfname1, imfname2, ...
  mincnt, px, dy1, dy2)
% [IM1, IM2, PM, PR] = PAYOFFMATRIXIM(ARG1, ARG2, IMFNAME1, IMFNAME2,
%   PX, MINCNT, DY1, DY2)
%
% Create a color coded Payoff Matrix images using data created
% by payoffmatrix.pl. First image is pair-pair percentage; second
% is pair-pair over/under-performance.
%
% ARG1     - Payoff Matrix structure or filename of tab delimited file
%            created by payoffmatrix.pl (using -pm switch)
% ARG2     - Pair structure or filename of tab delimited pairs
%            created by payoffmatrix.pl (using -pr switch)
% IMFNAME1 - (Optional) Image filename for pair-pair percentage
% IMFNAME2 - (Optional) Image filename for over/under-performance
% PX       - (Optional) Size of interaction cell (in pixels). Default is 20.
% MINCNT   - (Optional) Minimum number of boards on which result must be
%            based to avoid flagging. Default is 10.
% DY1      - Dynamic range of the color scale for pair-pair percentage.
%            Default is 30, meaning 50-30:50+30, i.e. 20-80%
% DY1      - Dynamic range of the color scale for under/over-performace.
%            Default is 60, meaning -30%:30%
%
% IM1      - Pair-pair percentage image
% IM2      - Pair-pair over/under-performance image
% PM       - Payoff matrix data (useful if read from a file)
% PR        -Pair data (useful if read from a file)

% This software is released under the GNU General Public License GPLv3
% See: http://www.gnu.org/licenses/gpl.html for full license.
%
% 05-Oct-2013 - Last modification
% 01-Oct-2013 - Original code by Matthew Kidd (San Diego, CA)

DEFAULT_PX = 10;
DEFAULT_DY1 = 30;
DEFAULT_DY2 = 60;
DEFAULT_MINCNT = 10;

NO_DATA_COLOR   = [255 255 255];
CANT_MEET_COLOR = [255 128 128];

error(nargchk(2,8,nargin));

if ~exist('px', 'var') || isempty(px), px = DEFAULT_PX; end
if ~exist('dy1', 'var') || isempty(dy1), dy1 = DEFAULT_DY1; end
if ~exist('dy2', 'var') || isempty(dy2), dy2 = DEFAULT_DY2; end
if ~exist('mincnt', 'var') || isempty(mincnt), mincnt = DEFAULT_MINCNT; end

if exist('imfname1', 'var') && ~(ischar(imfname1) || isempty(imfname1))
  error('IMFNAME1 must be a character string or [].')
end
if exist('imfname2', 'var') && ~(ischar(imfname2) || isempty(imfname2))
  error('IMFNAME2 must be a character string or [].')
end

if ischar(arg1)
  fd = fopen(arg1, 'r');
  if (fd == -1), error('Failed to open/read: %s', arg1); end
  cols = textscan(fd, '%d%d%*s%*s%*s%*s%f%f%f%f%f%f', 'Delimiter', '\t', ...
    'Headerlines', 1);
  fclose(fd);
  pm = cell2struct(cols, ...
    {'ix1', 'ix2', 'cnt', 'pct', 'meanerr', 'std', 'disparity', 'relperf'}, 2);
elseif ~isstruct(arg1)
  error('First argument must be a character array or a struct.');
else
  pm = arg1;
end

if ischar(arg2)
  fd = fopen(arg2, 'r');
  if (fd == -1), error('Failed to open/read: %s', arg2); end
  cols = textscan(fd, '%*d%d%d%*[^\n]', 'Delimiter', '\t', 'Headerlines', 1);
  fclose(fd);
  pr = cell2struct(cols, {'pnum1', 'pnum2'}, 2);
elseif ~isstruct(arg2)
  error('Second argument must be a character array or a struct.');
else
  pr = arg2;
end

% Figure out which pairs can never meet because they share a player in
% common. This isn't the most space efficient approach but it will be fast.
npairs = length(pr.pnum1);
p1x = repmat(pr.pnum1,  [1 npairs]);
p1y = repmat(pr.pnum1', [npairs 1]);
p2x = repmat(pr.pnum2,  [1 npairs]);
p2y = repmat(pr.pnum2', [npairs 1]);
om = p1x == p1y | p1x == p2y | p2x == p1y | p2x == p2y;

% Prepare the aspects of the image that are common to both output images.
im = zeros(npairs*px, npairs*px, 3, 'uint8');
for ic = 1:3, im(:,:,ic) = NO_DATA_COLOR(ic); end
for ix = 1:npairs
  for iy = 1:npairs
    for ic = 1:3
      if om(ix,iy)
        im((ix-1)*px+1:ix*px, (iy-1)*px+1:iy*px, ic) = CANT_MEET_COLOR(ic);
      end
    end
  end
end

im1 = fillImage(im, pm, 'pct', px, 50, dy1, mincnt);
figure
imshow(im1);
im2 = fillImage(im, pm, 'relperf', px, 0, dy2, mincnt);
figure
imshow(im2);

if exist('imfname1', 'var') && ~isempty(imfname1)
  imwrite(im1, imfname1, 'png');
end
if exist('imfname2', 'var') && ~isempty(imfname1)
  imwrite(im2, imfname2, 'png');
end

if nargout > 0, varargout{1} = im1; end
if nargout > 1, varargout{2} = im2; end
if nargout > 2, varargout{3} = pm; end
if nargout > 3, varargout{4} = pr; end


function im = fillImage(im, pm, fdname, px, mid, dy, mincnt)
% PM     - Structure
% FDNAME - Field name of pm containing the data to use
% PX     - Number of pixels hide and wide for each cell
% MID    - Middle value of the range
% DY     - Dynamic range. MID-DY and MID+DY are top and bottom of the color
%          range
% MINCNT - Interactions based on fewer than MINCNT hands will be stippled.

POS_COLOR = [255 255 0];  % yellow
NEG_COLOR = [0 0 255];    % blue
MID_COLOR = [0 0 0];
X_COLOR_R  = 192;
X_COLOR_GB = 128;

LOW_CNT_INDICATOR = 'reddishX';
% LOW_CNT_INDICATOR = 'checker';

% Create PX x PX indices for 'checker' option.
ci1 = 0:2:px-1;
ci2 = 1:2:px-1;
ci = zeros(round(px/2)^2 + floor(px/2)^2, 1, 'int32');
for k=0:round(px/2)-1, ci(k*px+1:k*px+round(px/2)) = ci1 + size(im,1)*2*k; end
for k=0:floor(px/2)-1, ci(k*px+round(px/2)+1:(k+1)*px) = ci2 + size(im,1)*(2*k+1); end

data = pm.(fdname);
nelem = length(pm.ix1);
for i=1:nelem
  v = (data(i) - mid) / dy;
  if v>1, v=1; elseif v<-1, v=-1; end
    
  iy = pm.ix1(i);
  ix = pm.ix2(i);
  if (iy == 35 && ix == 1)
    fprintf(1, '%0.2f %0.2f %0.2f\n', v, pm.pct(i), pm.relperf(i));
  end
  
  if (0)
    % Old strategy. Transition around midpoint flip too harshly.
    
    % Avoid darkest colors because it is hard to tell the difference
    % between the dark shades.
    v = sign(v) * (0.4 + abs(v)/2);
    
    for ic = 1:3
      if v > 0
        im( (iy-1)*px+1:iy*px, (ix-1)*px+1:ix*px, ic ) =  v * POS_COLOR(ic);
        im( (ix-1)*px+1:ix*px, (iy-1)*px+1:iy*px, ic ) =  v * NEG_COLOR(ic);
      elseif v < 0
        im( (iy-1)*px+1:iy*px, (ix-1)*px+1:ix*px, ic ) = -v * NEG_COLOR(ic);
        im( (ix-1)*px+1:ix*px, (iy-1)*px+1:iy*px, ic ) = -v * POS_COLOR(ic);
      else
        im( (iy-1)*px+1:iy*px, (ix-1)*px+1:ix*px, ic ) = MID_COLOR(ic);
        im( (ix-1)*px+1:ix*px, (iy-1)*px+1:iy*px, ic ) = MID_COLOR(ic);
      end
    end
  else
    % Try blending strategy.
    mb = 0.0;
    frac = (v+1)/2;
    blend1 = (mb + (1-mb)*frac) * POS_COLOR + (mb + (1-mb)*(1-frac)) * NEG_COLOR;
    blend2 = (mb + (1-mb)*(1-frac)) * POS_COLOR + (mb + (1-mb)*frac) * NEG_COLOR;
    
    for ic = 1:3
      im( (iy-1)*px+1:iy*px, (ix-1)*px+1:ix*px, ic ) =  blend1(ic);
      im( (ix-1)*px+1:ix*px, (iy-1)*px+1:iy*px, ic ) =  blend2(ic);
    end
  end

  % Flag matrix elements based on low statistics
  if pm.cnt(i) < mincnt && pm.cnt(i) ~= 0;
    switch LOW_CNT_INDICATOR
      case 'reddishX'
        % Add white X.
        for kd = 3:px-2
          im( (iy-1)*px+kd, (ix-1)*px+kd, 1) = X_COLOR_R;
          im( (ix-1)*px+kd, (iy-1)*px+kd, 1) = X_COLOR_R;
          im( (iy-1)*px+kd, ix*px+1-kd, 1)   = X_COLOR_R;
          im( (ix-1)*px+kd, iy*px+1-kd, 1)   = X_COLOR_R;

          im( (iy-1)*px+kd, (ix-1)*px+kd, 2:3) = X_COLOR_GB;
          im( (ix-1)*px+kd, (iy-1)*px+kd, 2:3) = X_COLOR_GB;
          im( (iy-1)*px+kd, ix*px+1-kd, 2:3)   = X_COLOR_GB;
          im( (ix-1)*px+kd, iy*px+1-kd, 2:3)   = X_COLOR_GB;        
        end
        
      case 'checker'
        im( px * (iy + size(im,1)*(ix-1)) + 1 + ci ) = 255;
    end

  end
end








  